from contextlib import asynccontextmanager
from typing import Any, Awaitable, Callable, List, Dict, Optional, Union

import anyio
import anyio.to_thread

from aas_middleware.connect.workflows.workflow import Workflow, typechecked_partial


@asynccontextmanager
async def blocking_aquire(semaphore: anyio.Semaphore):
    if semaphore.value == 0:
        raise RuntimeError("Maximum number of Workflows already running.")
    async with semaphore:
        yield

class BlockingWorkflow(Workflow):
    """
    The blocking workflow class is a subclass of the Workflow class and is used to define workflows that are blocking concurrent executions.
    The number of concurrently running workflows is limited to the pool size number. If another workflow is executed, its start is aborted.

    Args:
        workflow_function (Union[Awaitable[None], Callable[..., None]]): The workflow function to be executed.
        interval (Optional[float]): The interval in seconds in which the workflow function is executed.
        on_startup (bool, optional): If True, the workflow function is executed on startup. Defaults to False.
        on_shutdown (bool, optional): If True, the workflow function is executed on shutdown. Defaults to False.
        pool_size (int, optional): The number of concurrently running workflows. Defaults to 1.
    """
    def __init__(
        self,
        workflow_function: Union[Awaitable[None], Callable[..., None]],
        interval: Optional[float],
        on_startup: bool = False,
        on_shutdown: bool = False,
        pool_size: int = 1,
    ):
        super().__init__(workflow_function, interval, on_startup, on_shutdown)
        self.pool_size = pool_size
        self.semaphore = anyio.Semaphore(pool_size)

    @classmethod
    def define(
        cls,
        func: Awaitable,
        *args: List[Any],
        on_startup: bool,
        on_shutdown: bool,
        interval: Optional[float],
        pool_size: int=1,
        **kwargs: Dict[str, Any],
    ):
        workflow_function = typechecked_partial(func, *args, **kwargs)
        return cls(
            workflow_function=workflow_function,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            interval=interval,
            pool_size=pool_size,
        )

    async def _run_workflow_function(self, *args, **kwargs) -> Awaitable[Any]:
        async with blocking_aquire(self.semaphore):
            return await super()._run_workflow_function(*args, **kwargs)


from typing import Any, Awaitable, Callable, List, Dict, Optional, Union

import anyio
import anyio.to_thread


from aas_middleware.connect.workflows.workflow import Workflow, typechecked_partial

class QueueingWorkflow(Workflow):
    """
    The queueing workflow class is a subclass of the Workflow class and is used to define workflows that are executed concurrently. The number of concurrently running workflows is limited to the pool size number. If another workflow is executed, it is queued until a slot is available.

    Args:
        workflow_function (Union[Awaitable[None], Callable[..., None]]): The workflow function to be executed.
        interval (Optional[float]): The interval in seconds in which the workflow function is executed.
        on_startup (bool, optional): If True, the workflow function is executed on startup. Defaults to False.
        on_shutdown (bool, optional): If True, the workflow function is executed on shutdown. Defaults to False.
        pool_size (int, optional): The number of concurrently running workflows. Defaults to 1.
    """
    def __init__(
        self,
        workflow_function: Union[Awaitable[None], Callable[..., None]],
        interval: Optional[float],
        on_startup: bool = False,
        on_shutdown: bool = False,
        pool_size: int = 1,
    ):
        super().__init__(workflow_function, interval, on_startup, on_shutdown)
        self.pool_size = pool_size
        self.semaphore = anyio.Semaphore(pool_size)

    @classmethod
    def define(
        cls,
        func: Awaitable,
        *args: List[Any],
        on_startup: bool,
        on_shutdown: bool,
        interval: Optional[float],
        pool_size: int=1,
        **kwargs: Dict[str, Any],
    ):
        workflow_function = typechecked_partial(func, *args, **kwargs)
        return cls(
            workflow_function=workflow_function,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            interval=interval,
            pool_size=pool_size,
        )

    async def _run_workflow_function(self, *args, **kwargs) -> Awaitable[Any]:
        async with self.semaphore:
            return await super()._run_workflow_function(*args, **kwargs)



import inspect
from typing import Any, Awaitable, Callable, List, Dict, Optional, Union
import functools
import uuid

import anyio
from anyio.abc import TaskGroup
import anyio.to_thread
from exceptiongroup import ExceptionGroup, catch
import typeguard

from aas_middleware.connect.workflows.worfklow_description import WorkflowDescription

def typechecked_partial(func, *args, **kwargs):
    wrapped = typeguard.typechecked(func)
    return functools.partial(wrapped, *args, **kwargs)

class Workflow:
    """
    The Workflow class is used to define workflows that are executed. It allows to define a workflow function that is executed once or repeatedly. Also, execution of the workflow can be interrupted.

    Args:
        workflow_function (Union[Awaitable[None], Callable[..., None]]): The workflow function to be executed
        interval (Optional[float]): The interval in seconds in which the workflow function is executed
        on_startup (bool, optional): If True, the workflow function is executed on startup. Defaults to False.
        on_shutdown (bool, optional): If True, the workflow function is executed on shutdown. Defaults to False.
    """
    def __init__(
        self,
        workflow_function: Union[Awaitable[None], Callable[..., None]],
        interval: Optional[float],
        on_startup: bool = False,
        on_shutdown: bool = False,
    ):
        if not isinstance(workflow_function, functools.partial):
            workflow_function = typeguard.typechecked(workflow_function)
        self.workflow_function = workflow_function
        self.on_startup = on_startup
        self.on_shutdown = on_shutdown
        self.interval = interval
        self.task_groups: Dict[str, TaskGroup] = {}

    @property
    def running(self) -> bool:
        return len(self.task_groups) > 0

    def get_name(self) -> str:
        if self.workflow_function is None:
            raise ValueError(
                "No workflow function defined. Use the 'define' method to define a workflow function."
            )
        return self.workflow_function.func.__qualname__.split(".")[-1]

    def get_description(self) -> WorkflowDescription:
        return WorkflowDescription(
            name=self.get_name(),
            running=self.running,
            on_startup=self.on_startup,
            on_shutdown=self.on_shutdown,
            interval=self.interval,
            consumers=[],
            providers=[],
        )

    @classmethod
    def define(
        cls,
        func: Awaitable,
        *args: List[Any],
        on_startup: bool,
        on_shutdown: bool,
        interval: Optional[float],
        **kwargs: Dict[str, Any],
    ):
        workflow_function = typechecked_partial(
            func, *args, **kwargs
        )
        return cls(
            workflow_function=workflow_function,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            interval=interval
        )

    async def _run_workflow_function(self, *args, **kwargs) -> Awaitable[Any]:
        if inspect.iscoroutinefunction(self.workflow_function):
            return await self.workflow_function(*args, **kwargs)
        else:
            partial_func = functools.partial(
                self.workflow_function, *args, **kwargs
            )
            return await anyio.to_thread.run_sync(
                partial_func, abandon_on_cancel=True
            )

    def handle_error(self, excgroup: ExceptionGroup, execution_id: str) -> None:
        for exc in excgroup.exceptions:
            if execution_id in self.task_groups:
                del self.task_groups[execution_id]
            raise RuntimeError(f"Error during execution of workflow: {str(exc)}".replace("\"", "'"))

    async def execute(self, *args, **kwargs) -> Awaitable[Any]:
        """
        Args:
            *args: list of arguments
            **kwargs: dictionary of keyword arguments

        Raises:
            RuntimeError: Error during execution of workflow

        Returns:
            Awaitable[Any]: return value of the workflow function
        """
        if not self.interval or self.interval == 0.0:
            return_value = await self._execute_once(*args, **kwargs)
            return return_value
        else:
            await self._execute_repeatedly(*args, **kwargs)

    async def _execute_once(self, *args, **kwargs) -> Awaitable[None]:
        execution_id = str(uuid.uuid4())
        handle_error_partial = functools.partial(self.handle_error, execution_id=execution_id)
        with catch({Exception: handle_error_partial}):
            async with anyio.create_task_group() as tg:
                self.task_groups[execution_id] = tg
                return_value = await self._run_workflow_function(*args, **kwargs)
                del self.task_groups[execution_id]
                return return_value
            if execution_id in self.task_groups:
                del self.task_groups[execution_id]
                raise RuntimeError("Workflow was interrupted.")

    async def _execute_repeatedly(self, *args, **kwargs) -> Awaitable[None]:
        execution_id = str(uuid.uuid4())
        handle_error_partial = functools.partial(self.handle_error, execution_id=execution_id)
        with catch({Exception: handle_error_partial}):
            async with anyio.create_task_group() as tg:
                self.task_groups[execution_id] = tg
                while True:
                    await self._run_workflow_function(*args, **kwargs)
                    await anyio.sleep(self.interval)
            del self.task_groups[execution_id]

    async def interrupt(self):
        if not self.running:
            raise ValueError("No workflow is running.")
        for _, task_group in self.task_groups.items():
            if task_group.cancel_scope.cancel_called or task_group.cancel_scope.cancelled_caught:
                continue
            task_group.cancel_scope.cancel()



import functools
from inspect import signature
from typing import Dict, List
import typing

from fastapi import APIRouter, BackgroundTasks, HTTPException
from pydantic import BaseModel, Field, create_model
from aas_middleware.connect.workflows.worfklow_description import WorkflowDescription
from aas_middleware.connect.workflows.workflow import Workflow


def get_partial_type_hints(partial_func):
    original_func = partial_func.func
    original_hints = typing.get_type_hints(original_func)
    
    sig = signature(original_func)
    params = sig.parameters
    
    fixed_args = partial_func.args
    fixed_keywords = partial_func.keywords
    
    for index, param in enumerate(params.keys()):
        if index < len(fixed_args):
            original_hints.pop(param, None)
    
    for key in fixed_keywords.keys():
        original_hints.pop(key, None)
    
    return original_hints


def get_base_model_from_type_hints(name: str, type_hints: Dict[str, typing.Any]) -> typing.Type[BaseModel]:
    """
    Get the base model from the type hints of a function.

    Args:
        type_hints (Dict[str, typing.Any]): Type hints of a function.

    Returns:
        typing.Type[typing.Any]: Base model of the type hints.
    """
    if "return" in type_hints:
        type_hints.pop("return")
    dynamical_model_creation_dict = {}
    for argument, argument_type in type_hints.items():
        entry = {
            argument: typing.Annotated[
                argument_type, Field()
            ]
        }
        dynamical_model_creation_dict.update(entry)
    base_model = create_model(f"body_for_{name}", **dynamical_model_creation_dict)
    return base_model


def generate_workflow_endpoint(workflow: Workflow) -> List[APIRouter]:
    """
    Generates endpoints for a workflow to execute the workflow.

    Args:
        workflow (Workflow): Workflow that contains the function to be executed by the workflow.

    Returns:
        APIRouter: FastAPI router with an endpoint to execute the workflow.
    """
    router = APIRouter(
        prefix=f"/workflows/{workflow.get_name()}",
        tags=["workflows"],
        responses={404: {"description": "Not found"}},
    )

    if isinstance(workflow.workflow_function, functools.partial):
        type_hints = get_partial_type_hints(workflow.workflow_function)
    else:
        type_hints = typing.get_type_hints(workflow.workflow_function)
    
    return_type = type_hints.pop("return") if "return" in type_hints else None
    if len(type_hints) == 0:
        input_type_hints = None
    elif len(type_hints) == 1:
        input_type_hints = list(type_hints.values())[0]
    else:
        input_type_hints = get_base_model_from_type_hints(workflow.get_name(), type_hints)


    if input_type_hints is None:
        if workflow.get_description().interval is None:
            @router.post("/execute", response_model=return_type)
            async def execute():
                try:
                    return await workflow.execute()
                except RuntimeError as e:
                    raise HTTPException(status_code=409, detail=str(e))

        @router.post("/execute_background", response_model=Dict[str, str])
        async def execute_background(background_tasks: BackgroundTasks):
            background_tasks.add_task(workflow.execute)
            return {"message": f"Started exeuction of workflow {workflow.get_name()}"}
    else:
        if workflow.get_description().interval is None:
            @router.post("/execute", response_model=return_type)
            async def execute(arg: typing.Optional[input_type_hints]=None): # type: ignore
                try:
                    if isinstance(arg, BaseModel) and len(type_hints) > 1:
                        input_value = dict(arg)
                        return await workflow.execute(**input_value)
                    else:
                        return await workflow.execute(arg)
                except RuntimeError as e:  
                    raise HTTPException(status_code=409, detail=str(e))

        @router.post("/execute_background", response_model=Dict[str, str])
        async def execute_background(background_tasks: BackgroundTasks, arg: typing.Optional[input_type_hints]=None): # type: ignore
            if isinstance(arg, BaseModel) and len(type_hints) > 1:
                input_value = dict(arg)
                background_tasks.add_task(workflow.execute, **input_value)
            else:
                background_tasks.add_task(workflow.execute, arg)
            return {"message": f"Started exeuction of workflow {workflow.get_name()}"}
        

    @router.get("/description", response_model=WorkflowDescription)
    async def describe_workflow():
        return workflow.get_description()

    @router.get("/interrupt", response_model=Dict[str, str])
    async def interrupt_workflow():
        try:
            await workflow.interrupt()
        except ValueError as e:
            raise HTTPException(status_code=409, detail=str(e))
        return {"message": f"Stopped execution of workflow {workflow.get_name()}"}

    return router
